/*
 * A Unique LED Clock - Mood Lightƒ - Light Show using an Arduino Uno
 *
 * HARDWARE:
 * Processor board: Arduino Uno
 *
 *  For display: an LPD8806 LED strip containing 14 RGB LEDs. One LED for each hour
 *               plus two LEDs for misc usage. Available from adafruit.com.
 *               pin 4 is data to LEDs; pin 5 is clock to LEDs
 *
 *  For time keeping: a ChronoDot module based on the DS3231 RTC chip. Accurate to
 *                    one minute a year. Available from adafruit.com.
 *                    pin A4 is data to RTC; pin A5 is clock to RTC
 *
 *  For IR detection: Radio Shack IR Receiver #2760640 or Adafruit model TSOP38238 ID: 157
 *
 *  For remote control: Adafruit 21 key IR remote ID: 389
 *
 * Hardware and software design by: Craig A. Lindley
 * Code compiles on Arduino 1.0 or newer only
 *
 * Revision History
 *
 * Version: 1.0 - 10/31/2013
 *   Initial release
 */

#include <Arduino.h>
#include <avr/eeprom.h>
#include <IRremote.h>
#include <Wire.h>

#include "datatypes.h"

#define DEBUG false

// RGB data structure
struct RGB {
  uint8_t red;
  uint8_t green;
  uint8_t blue;
};

// HSV data structure
struct HSV {
  float hue;
  float sat;
  float val;
};

// EEPROM memory allocations
// First four bytes have signature
// 0xAA, 0x00, 0xFF, 0x55 if valid

// Allocations for mood light mode
#define MLM_HUE_ADDRESS             4
#define MLM_SAT_ADDRESS             8
#define MLM_VAL_ADDRESS            12

// Allocations for clock mode
// For clock background color
#define CLK_BACKGROUND_HUE_ADDRESS 16
#define CLK_BACKGROUND_SAT_ADDRESS 20
#define CLK_BACKGROUND_VAL_ADDRESS 24

// For clock minute color
#define CLK_MIN_HUE_ADDRESS        28
#define CLK_MIN_SAT_ADDRESS        32
#define CLK_MIN_VAL_ADDRESS        36

// For clock hour color
#define CLK_HOUR_HUE_ADDRESS       40
#define CLK_HOUR_SAT_ADDRESS       44
#define CLK_HOUR_VAL_ADDRESS       48

// For clock day color
#define CLK_DAY_HUE_ADDRESS        52
#define CLK_DAY_SAT_ADDRESS        56
#define CLK_DAY_VAL_ADDRESS        60

// For clock month color
#define CLK_MON_HUE_ADDRESS        64
#define CLK_MON_SAT_ADDRESS        68
#define CLK_MON_VAL_ADDRESS        72

// For clock year color
#define CLK_YEAR_HUE_ADDRESS       76
#define CLK_YEAR_SAT_ADDRESS       80
#define CLK_YEAR_VAL_ADDRESS       84

// For clock toggle color
#define CLK_TOGGLE_HUE_ADDRESS     88
#define CLK_TOGGLE_SAT_ADDRESS     92
#define CLK_TOGGLE_VAL_ADDRESS     96

// Arduino pin assignments
#define LED_PIN             13
#define LED_STRIP_CLOCK_PIN  5
#define LED_STRIP_DATA_PIN   4

/************************************************************************/
/************************* LED strip definitions ************************/
/************************************************************************/

// Count of total LEDs
#define NUMBER_OF_LEDS 14

// Masks and pointers to speed access
uint8_t clockPinMask, dataPinMask;
volatile uint8_t *clockPort, *dataPort;

// Array of LED data in GRB order. Component data is 7 bits wide.
uint32_t LEDData[NUMBER_OF_LEDS];

/************************************************************************/
/********************** Real time clock definitions *********************/
/************************************************************************/

#define DEFAULT_CENTURY 2000

#define CHRONODOT_I2C_ADDRESS    104
#define CHRONODOT_SECONDS_REGISTER 0
#define CHRONODOT_MINUTES_REGISTER 1
#define CHRONODOT_HOURS_REGISTER   2
#define CHRONODOT_DAY_REGISTER     3
#define CHRONODOT_DATE_REGISTER    4
#define CHRONODOT_MONTH_REGISTER   5
#define CHRONODOT_YEAR_REGISTER    6
#define CHRONODOT_STATUS_REGISTER 15

uint8_t seconds;  // 0..59
uint8_t minutes;  // 0..59
uint8_t hours;    // 1..12
uint8_t date;     // 1..31
uint8_t day;      // 1..7
uint8_t month;    // 1..12
uint8_t year;     // 0..99
uint8_t pmAm;     // 0 == AM : 1 == PM

boolean secondsToggle;
boolean hourMinuteToggle;
uint8_t latchedMinute;

/************************************************************************/
/******************* IR receiver and remote definitions *****************/
/************************************************************************/

// Pin IR receiver is connected to
int IR_RECV_PIN = 6;

// Declare instance of IR receiver
IRrecv irReceiver(IR_RECV_PIN);

// Result from decode operation
decode_results results;
unsigned long irCode;

// Keycodes for the Car IR remote control from Adafruit
#define CODE_VOLDOWN   0xFD00FF
#define CODE_PLAYPAUSE 0xFD807F
#define CODE_VOLUP     0xFD40BF
#define CODE_SETUP     0xFD20DF
#define CODE_UP        0xFDA05F
#define CODE_STOPMODE  0xFD609F
#define CODE_LEFT      0xFD10EF
#define CODE_ENTERSAVE 0xFD906F
#define CODE_RIGHT     0xFD50AF
#define CODE_010P      0xFD30CF
#define CODE_DOWN      0xFDB04F
#define CODE_BACK      0xFD708F
#define CODE_1         0xFD08F7
#define CODE_2         0xFD8877
#define CODE_3         0xFD48B7
#define CODE_4         0xFD28D7
#define CODE_5         0xFDA857
#define CODE_6         0xFD6897
#define CODE_7         0xFD18E7
#define CODE_8         0xFD9867
#define CODE_9         0xFD58A7

/************************************************************************/
/*************************** LED strip functions ************************/
/************************************************************************/

// Write a byte of LED data to LED strip
void writeByte(byte b) {

  for (uint8_t bit = 0x80; bit; bit >>= 1) {      
    if (b & bit) {
      *dataPort |=  dataPinMask;
    }  
    else {
      *dataPort &= ~dataPinMask;
    }
    *clockPort |=  clockPinMask;
    *clockPort &= ~clockPinMask;
  }
}

// Push LED data to LED strip
void show() {

  // Get the strip's attention
  writeByte(0);

  // Write 24 bits/LED
  for (int i = 0; i < NUMBER_OF_LEDS; i++) {
    uint32_t pixel = LEDData[i];
    writeByte(pixel >> 16 & 0xFF);
    writeByte(pixel >> 8  & 0xFF);
    writeByte(pixel       & 0xFF);
  }
  // Finish up
  writeByte(0);
}

// Brightness correction data for LEDs.
// Used for dimming LEDs uniformly
uint8_t GAMMA [] = {
  0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,
  1,2,2,2,2,2,2,3,3,3,4,4,4,5,5,6,
  6,7,7,8,9,9,10,11,12,13,14,16,17,18,20,21,
  23,25,27,29,31,33,36,38,41,43,46,49,52,55,58,61,
  64,66,69,72,75,78,81,84,86,89,91,94,96,98,100,102,
  104,106,107,109,110,111,113,114,115,116,117,118,118,119,120,120,
  121,121,122,122,123,123,123,124,124,124,125,125,125,125,125,125,
  126,126,126,126,126,126,126,126,126,126,126,126,127,127,127,127
};

// Convert separate R,G,B values into packed 32-bit color
uint32_t createPackedColor(uint8_t r, uint8_t g, uint8_t b) {

  // Notice the order of the color components is GRB not RGB
  return ((uint32_t) g << 16) | ((uint32_t) r << 8) | (uint32_t) b;
}

// Set LED color from separate 7-bit R, G, B components
void setLEDColor(uint16_t ledNumber, uint8_t r, uint8_t g, uint8_t b) {
  uint32_t data = (g | 0x80);
  data <<= 8;
  data |= (r | 0x80);
  data <<= 8;
  data |= (b | 0x80);
  LEDData[ledNumber] = data;
}

// Set LED color from RGB struct
void setLEDColor(uint16_t ledNumber, struct RGB color) {
  setLEDColor(ledNumber, color.red, color.green, color.blue);  
}

// Set LED color from packed RGB value
void setLEDColor(uint16_t ledNumber, uint32_t c) {

  uint8_t g = ((c >> 16) & 0x7F) | 0x80;
  uint8_t r = ((c >>  8) & 0x7F) | 0x80;
  uint8_t b = ((c)       & 0x7F) | 0x80;

  uint32_t data = g;
  data <<= 8;
  data |= r;
  data <<= 8;
  data |= b;
  LEDData[ledNumber] = data;
}

// Set all LEDs to black
void clearLEDs() {
  for (int i = 0; i < NUMBER_OF_LEDS; i++) {
    setLEDColor(i, 0);
  }
  show();
}

/************************************************************************/
/*********************** Color creation functions ***********************/
/************************************************************************/

// Color definitions
#define MIN_LED_VALUE   0 
#define MAX_LED_VALUE 127

// Color space conversion
// Input arguments
// hue in degrees (0 - 360.0)
// saturation in percent (0.0 - 1.0)
// value in percent (0.0 - 1.0)
// Output arguments
// red, green blue (0.0 - 1.0)
void _HSVtoRGB(double hue, double saturation, double value, double *red, double *green, double *blue) {

  int i;
  double f, p, q, t;
  if (saturation == 0) {
    // achromatic (grey)
    *red = *green = *blue = value;
    return;
  }
  hue /= 60;			// sector 0 to 5
  i = floor(hue);
  f = hue - i;			// factorial part of h
  p = value * (1 - saturation);
  q = value * (1 - saturation * f);
  t = value * (1 - saturation * (1 - f));
  switch(i) {
  case 0:
    *red = value;
    *green = t;
    *blue = p;
    break;
  case 1:
    *red = q;
    *green = value;
    *blue = p;
    break;
  case 2:
    *red = p;
    *green = value;
    *blue = t;
    break;
  case 3:
    *red = p;
    *green = q;
    *blue = value;
    break;
  case 4:
    *red = t;
    *green = p;
    *blue = value;
    break;
  default:
    *red = value;
    *green = p;
    *blue = q;
    break;
  }
}

// Create color. Store in RGB structure
void createColor(float hue, float saturation, float value, struct RGB *color) {
  double r, g, b;

  _HSVtoRGB(hue, saturation, value, &r, &g, &b);

  (*color).red   = (uint8_t) (r * MAX_LED_VALUE);
  (*color).green = (uint8_t) (g * MAX_LED_VALUE);
  (*color).blue =  (uint8_t) (b * MAX_LED_VALUE);
}

// Create a fully saturated HSV color
void createHSVColor(byte divisions, byte index, struct RGB *color) {

  float hueAngle = (360.0 * index) / divisions;
  createColor(hueAngle, 1.0, 1.0, color);
}

// Default clock colors
// Default background color blue
uint32_t DEFAULT_BACKGROUND_COLOR = createPackedColor(MIN_LED_VALUE, MIN_LED_VALUE, MAX_LED_VALUE);

// Default minute color green
uint32_t DEFAULT_MIN_COLOR = createPackedColor(MIN_LED_VALUE, MAX_LED_VALUE, MIN_LED_VALUE);

// Default hour color red
uint32_t DEFAULT_HOUR_COLOR = createPackedColor(MAX_LED_VALUE, MIN_LED_VALUE, MIN_LED_VALUE);

// Default day color magenta
uint32_t DEFAULT_DAY_COLOR = createPackedColor(MAX_LED_VALUE, MIN_LED_VALUE, MAX_LED_VALUE);

// Default month color yellow
uint32_t DEFAULT_MON_COLOR = createPackedColor(MAX_LED_VALUE, MAX_LED_VALUE, MIN_LED_VALUE);

// Default year color cyan
uint32_t DEFAULT_YEAR_COLOR = createPackedColor(MIN_LED_VALUE, MAX_LED_VALUE, MAX_LED_VALUE);

// Default toggle color red
uint32_t DEFAULT_TOGGLE_COLOR = createPackedColor(MAX_LED_VALUE, MIN_LED_VALUE, MIN_LED_VALUE);

// Allow manual adjustment of hue, saturation and value from the remote control
// Call out is made to callback fn to display configured color. On return, the HSV
// structure has any changes to the color parameters the user made.
void hsvColorSetter(struct HSV *hsv, callback fn) {

  // This process has two distinct states
  enum CS_STATES {
    STATE_WAITING, STATE_SHOW_COLOR
  };

  const float hueIncrement = 10.0;  // 36 possible hue values
  const float satIncrement =  0.1;  // 10 possible saturation values
  const float valIncrement =  0.05; // 20 possible value values

  clearLEDs();

  // Set initial state
  CS_STATES csState = STATE_SHOW_COLOR;

  while (true) {

    switch (csState) {
    case STATE_WAITING:
      {
        // Read the remote
        irCode = readIRCode();

        while (irCode == 0) {
          delay(200);
          irCode = readIRCode();
        }

        // Process the code
        switch (irCode) {

        case CODE_STOPMODE:
        case CODE_ENTERSAVE:
          // Return to caller with HSV structure possibly modified
          return;

        case CODE_1:  // Hue angle increment
          {
            hsv->hue += hueIncrement;
            if (hsv->hue >= 360) {
              hsv->hue = 0;
            }
            csState = STATE_SHOW_COLOR;
          }
          break;

        case CODE_7:
          {
            hsv->hue -= hueIncrement;
            if (hsv->hue < 0) {
              hsv->hue = 359;
            }
            csState = STATE_SHOW_COLOR;
          }
          break;

        case CODE_2:
          {
            hsv->sat += satIncrement;
            if (hsv->sat > 1.0) {
              hsv->sat = 1.0;
            }
            csState = STATE_SHOW_COLOR;
          }
          break;

        case CODE_8:
          {
            hsv->sat -= satIncrement;
            if (hsv->sat < 0.0) {
              hsv->sat = 0.0;
            }
            csState = STATE_SHOW_COLOR;
          }
          break;

        case CODE_3:
          {
            hsv->val += valIncrement;
            if (hsv->val > 1.0) {
              hsv->val = 1.0;
            }
            csState = STATE_SHOW_COLOR;
          }
          break;

        case CODE_9:
          {
            hsv->val -= valIncrement;
            if (hsv->val < 0.05) {
              hsv->val = 0.05;
            }
            csState = STATE_SHOW_COLOR;
          }
          break;
        }
      }
      break;

    case STATE_SHOW_COLOR:
      {
        // Do callback to display set color
        fn(hsv);

        // Set next state
        csState = STATE_WAITING;
      }
      break;
    }
  }
}

/************************************************************************/
/*********************** Chronodot Access Functions *********************/
/************************************************************************/

byte decToBcd(byte num) {
  return (((num / 10) << 4) | (num % 10));
}

byte bcdToDec(byte num) {
  return (((num >> 4) * 10) + (num & 0x0F));
}

// Determine if the Chronodot is running by examining the status register.
// If bit 7 is high, the oscillator is stopped which is bad
// Returns true if RTC is running; false otherwise
boolean rtcRunning() {

  // First reset the status register
  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_STATUS_REGISTER);
  Wire.write(0);
  Wire.endTransmission();

  // Now read back to see if osc is running
  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_STATUS_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 1);
  uint8_t status = Wire.read();

  return (status & (1 << 7)) == 0;
}

// Set the RTC into 12 hour format
void set12HourFormat() {

  // Read the hours register
  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_HOURS_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 1);
  uint8_t temp = Wire.read();

  temp |= (1 << 6);  // Set bit to enable 12 hour format

  // Write the modified hours value back to the RTC   
  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_HOURS_REGISTER);
  Wire.write(temp);
  Wire.endTransmission();
}

// Set the time to the RTC in 12 hour format
void setTime() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_SECONDS_REGISTER);
  Wire.write(decToBcd(seconds));
  Wire.write(decToBcd(minutes));

  // Process hours value for 12 hour format
  uint8_t data = decToBcd(hours);

  // Is the time in the PM ?
  if (pmAm) {
    // Yes, so set the Pm/Am bit
    data |= (1 << 5);
  }  
  else  {
    // No, so reset the Pm/Am bit
    data &= ~(1 << 5);    
  }
  // Finally set the 12/24 format bit 
  data |= (1 << 6);

  // Write the modified hours value
  Wire.write(data);
  Wire.endTransmission();
}

// Get the time from the RTC in 12 hour format
void getTime() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write((byte) CHRONODOT_SECONDS_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 3);
  seconds = bcdToDec(Wire.read() & 0x7F);
  minutes = bcdToDec(Wire.read() & 0x7F);

  // Read the hours value
  uint8_t temp = Wire.read();

  // Is it AM or PM
  pmAm = ((temp & (1 << 5)) != 0);

  // Mask off 12/24 and Pm/Am bits to get hours value
  hours = bcdToDec(temp & 0x1F);
} 

// Set the date to the RTC. day is day of the week and is ununsed
void setDate() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_DAY_REGISTER);
  Wire.write(decToBcd(day));
  Wire.write(decToBcd(date));
  Wire.write(decToBcd(month));
  Wire.write(decToBcd(year));
  Wire.endTransmission();
}

// Get the date from the RTC. day is day of the week and is ununsed
void getDate() {

  Wire.beginTransmission(CHRONODOT_I2C_ADDRESS);
  Wire.write(CHRONODOT_DAY_REGISTER);
  Wire.endTransmission();
  Wire.requestFrom(CHRONODOT_I2C_ADDRESS, 4);
  day   = bcdToDec(Wire.read());
  date  = bcdToDec(Wire.read());
  month = bcdToDec(Wire.read());
  year  = bcdToDec(Wire.read());
}

// Print out the current time from the RTC
void printDateTime() {  

  if(hours < 10) {
    Serial.print("0");
  }
  Serial.print(hours, DEC);
  Serial.print(':');

  if(minutes < 10) {
    Serial.print("0");
  }
  Serial.print(minutes, DEC);
  Serial.print(':');

  if(seconds < 10) {
    Serial.print("0");
  }
  Serial.print(seconds, DEC);

  Serial.print(' ');

  if (pmAm) {
    Serial.print("PM");
  }  
  else  {
    Serial.print("AM");
  }

  Serial.print(' ');

  if(month < 10) {
    Serial.print("0");
  }
  Serial.print(month, DEC);
  Serial.print('/');

  if(date < 10) {
    Serial.print("0");
  }
  Serial.print(date, DEC);
  Serial.print('/');

  Serial.print(year + DEFAULT_CENTURY, DEC);
  Serial.println();
}

/************************************************************************/
/****************************** IR Functions ****************************/
/************************************************************************/

// Low level IR code reading function
// Function will return 0 if no IR code available
unsigned long _readIRCode() {

  results.value = 0;

  // Attempt to read an IR code ?
  if (irReceiver.decode(&results)) {

    delay(20);

    // Prepare to receive the next IR code
    irReceiver.resume();
  }
  return results.value;
}

// Read an IR code
// Function will return 0 if no IR code available
unsigned long readIRCode() {

  // Is there an IR code to read ?
  unsigned long code = _readIRCode();
  if (code == 0) {
    // No code so return 0
    return 0;
  }
  // Keep reading until code changes
  while (_readIRCode() == code) {
    ;
  }
  return code;
}

/************************************************************************/
/********************** EEPROM read/write functions *********************/
/************************************************************************/

// Read uint8_t from the eeprom @ address
uint8_t eepromReadUint8_t(int address) {

  uint8_t value;

  eeprom_read_block(&value, (void *) address, sizeof(uint8_t));

  return value;
}

// Write uint8_t to the eeprom @ address
void eepromWriteUint8_t(int address, uint8_t value) {

  eeprom_write_block(&value, (void *) address, sizeof(uint8_t));
}

// Read uint16_t from the eeprom @ address
uint16_t eepromReadUint16_t(int address) {

  uint16_t value;

  eeprom_read_block(&value, (void *) address, sizeof(uint16_t));

  return value;
}

// Write uint16_t to the eeprom @ address
void eepromWriteUint16_t(int address, uint16_t value) {

  eeprom_write_block(&value, (void *) address, sizeof(uint16_t));
}

// Read uint32_t from the eeprom @ address
uint32_t eepromReadUint32_t(int address) {

  uint32_t value;

  eeprom_read_block(&value, (void *) address, sizeof(uint32_t));

  return value;
}

// Write uint32_t to the eeprom @ address
void eepromWrite32Int(int address, uint32_t value) {

  eeprom_write_block(&value, (void *) address, sizeof(uint32_t));
}

// Read float from the eeprom @ address
float eepromReadFloat(int address) {

  float value;

  eeprom_read_block((void *) &value, (void *) address, sizeof(float));

  return value;
}

// Write float to the eeprom @ address
void eepromWriteFloat(int address, float value) {

  eeprom_write_block((void *) &value, (void *) address, sizeof(float));
}

/************************************************************************/
/*************************** Reset EEPROM data **************************/
/************************************************************************/

void resetEEPROMData() {

  // Write the signature into the eeprom
  eepromWriteUint8_t(0, 0xAA);
  eepromWriteUint8_t(1, 0x00);
  eepromWriteUint8_t(2, 0xFF);
  eepromWriteUint8_t(3, 0x55);

  // All HSV color values to zero which will cause defaults to be used
  eepromWriteFloat(MLM_HUE_ADDRESS, 0);
  eepromWriteFloat(MLM_SAT_ADDRESS, 0);
  eepromWriteFloat(MLM_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_BACKGROUND_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_BACKGROUND_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_BACKGROUND_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_MIN_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_MIN_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_MIN_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_HOUR_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_HOUR_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_HOUR_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_DAY_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_DAY_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_DAY_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_MON_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_MON_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_MON_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_YEAR_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_YEAR_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_YEAR_VAL_ADDRESS, 0);

  eepromWriteFloat(CLK_TOGGLE_HUE_ADDRESS, 0);
  eepromWriteFloat(CLK_TOGGLE_SAT_ADDRESS, 0);
  eepromWriteFloat(CLK_TOGGLE_VAL_ADDRESS, 0);
}

/************************************************************************/
/*************************** Mood Light Mode ****************************/
/************************************************************************/

// Mood Light Mode callback
void mlmCallback(struct HSV *hsv) {

  struct RGB rgbColor;

  // Create color from the hsv data
  createColor(hsv->hue, hsv->sat, hsv->val, &rgbColor);

  // Set all LEDs to selected color
  for (int i = 0; i < NUMBER_OF_LEDS; i++) {
    setLEDColor(i, rgbColor);
  }
  show();
}

// Allow manual adjustment of hue, saturation and value
void moodLightMode() {

  struct HSV hsvColor;

  float hue, sat, val;

  // Read the values from the eeprom
  hue = eepromReadFloat(MLM_HUE_ADDRESS);
  sat = eepromReadFloat(MLM_SAT_ADDRESS);
  val = eepromReadFloat(MLM_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to gold
    hue = 30.0;
    sat = 1.0;
    val = 0.1;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &mlmCallback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(MLM_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(MLM_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(MLM_VAL_ADDRESS, hsvColor.val);
  }
}

/************************************************************************/
/*********************** Pattern Display Functions **********************/
/************************************************************************/

int multiplierIndex;
double delayMultiplier;

// Rotating color palette pattern
void pattern1() {

  RGB color;
  double hue;

  // Hue increment
  double hueInc = 360.0 / NUMBER_OF_LEDS - 2;

  // Repeat the pattern until interrupted
  while (true) {
    // Pattern rotates NUMBER_OF_LEDS times so each color has been in each position
    for (int s = 0; s < NUMBER_OF_LEDS - 2; s++) {
      for (int led = 2; led < NUMBER_OF_LEDS; led++) {
        hue = ((int)((led + s) * hueInc)) % 360;
        createColor(hue, 1.0, 0.5, &color);
        setLEDColor(led, color);
      }
      show();
      delay(250 * delayMultiplier);
      if (checkForTimeoutOrInterrupt()) {
        return;
      }
    }
  }
}

// Random color to random LED
void pattern2() {

  int led, r, g, b;

  // Repeat the pattern until interrupted
  while (true) {
    led = random(2, NUMBER_OF_LEDS);
    r = random(0, MAX_LED_VALUE);
    g = random(0, MAX_LED_VALUE);
    b = random(0, MAX_LED_VALUE);
    setLEDColor(led, r, g, b);
    show();
    delay(250 * delayMultiplier);
    if (checkForTimeoutOrInterrupt()) {
      return;
    }
  }
}

#define BRIGHTNESS_LEVELS 128
#define FADE_DELAY         20

// Fade in fade out display pattern
void fadeInFadeOutPattern(float hue) {

  RGB color;
  double value;

  while (true) {
    // Gradually increase color's value
    for (int b = 0; b < BRIGHTNESS_LEVELS; b++) {
      value = GAMMA[b] / 127.0;

      for (int led = 2; led < NUMBER_OF_LEDS; led++) {
        createColor(hue, 1.0, value, &color);
        setLEDColor(led, color);
      }
      show();    
      delay(FADE_DELAY * delayMultiplier);
      if (checkForTimeoutOrInterrupt()) {
        return;
      }
    } 

    // Gradually decrease color's value
    for (int b = BRIGHTNESS_LEVELS - 1 ; b >=0 ; b--) {
      value = GAMMA[b] / 127.0;

      for (int led = 2; led < NUMBER_OF_LEDS; led++) {
        createColor(hue, 1.0, value, &color);
        setLEDColor(led, color);
      }
      show();    
      delay(FADE_DELAY * delayMultiplier);
      if (checkForTimeoutOrInterrupt()) {
        return;
      }
    } 
  }
} 

// Red fade in fade out pattern
void pattern3() {
  fadeInFadeOutPattern(0);
}

// Green fade in fade out pattern
void pattern4() {
  fadeInFadeOutPattern(120);
}

// Blue fade in fade out pattern
void pattern5() {
  fadeInFadeOutPattern(240);
}

// Random color fade in fade out pattern
void pattern6() {  
  int hue = random(360);
  fadeInFadeOutPattern(hue);
}

// Rainbow color pattern
void pattern7() {

  uint8_t index, colorIndex = 0;
  RGB color;

  const uint8_t numberOfColors = 100;

  while (true) {
    index = colorIndex;
    for (byte led = 2; led < NUMBER_OF_LEDS; led++) {
      createHSVColor(numberOfColors, index, &color);
      setLEDColor(led, color);
      show();
      if (++index > numberOfColors - 1) {
        index = 0;
      }
    }
    if (checkForTimeoutOrInterrupt()) {
      return;
    }

    delay(200 * delayMultiplier);
    if (++colorIndex > numberOfColors - 1) {
      colorIndex = 0;
    }
  }
}

/************************************************************************/
/************************* Pattern Display Mode *************************/
/************************************************************************/

// The index of the next display pattern (if selected)
int selectorIndex;

// Delay multiplier constants and variables
#define MAX_MULTIPLIER_INDEX +4
#define MIN_MULTIPLIER_INDEX -4

// Random pattern timeout
#define PATTERN_DURATION_SECONDS 60

boolean timeOutEnabled;
unsigned long timeOut;

// States for the pattern selector
enum PDM_STATE {
  STATE_PATTERN_SELECT, STATE_PATTERN_EXECUTE, STATE_SPEED_UP, STATE_SPEED_DOWN, 
  STATE_PATTERN_NEXT, STATE_PATTERN_PREV, STATE_WAITING, STATE_EXITING
};

enum PDM_STATE pdmState;

// Array of pointers to pattern display functions
pt2Function patternFunctions [] = {

  pattern1,
  pattern2,
  pattern3,
  pattern4,
  pattern5,
  pattern6,
  pattern7,

  // Many more patterns could be added here
};

// Determine the number of display patterns from the entries in the array
#define NUMBER_OF_PATTERNS (sizeof(patternFunctions) / sizeof(pt2Function))

// Calculate the delay multiplier which will control the speed of
// the display pattern
void calculateDelayMultiplier(int theMultiplierIndex) {
  switch (theMultiplierIndex) {
  case 4:
    delayMultiplier = 16.0;
    break;
  case 3:
    delayMultiplier = 8.0;
    break;
  case 2:
    delayMultiplier = 4.0;
    break;
  case 1:
    delayMultiplier = 2.0;
    break;
  case 0:
    delayMultiplier = 1.0;
    break;
  case -1:
    delayMultiplier = 0.5;
    break;
  case -2:
    delayMultiplier = 0.25;
    break;
  case -3:
    delayMultiplier = 0.125;
    break;
  case -4:
    delayMultiplier = 0.0625;
    break;
  }
}

// Check for the pattern termination
// All patterns must call this periodically
boolean checkForTimeoutOrInterrupt() {

  // Check for user interrupt
  boolean interruptCondition = false;

  // Read the remote
  irCode = readIRCode();

  switch (irCode) {

  case CODE_STOPMODE:
    {
      pdmState = STATE_EXITING;
      interruptCondition = true;
    }
    break;

  case CODE_UP:
    {
      pdmState = STATE_SPEED_UP;
      interruptCondition = true;
    }
    break;

  case CODE_DOWN:
    {
      pdmState = STATE_SPEED_DOWN;
      interruptCondition = true;
    }
    break;

  case CODE_LEFT:
    {
      pdmState = STATE_PATTERN_PREV;
      interruptCondition = true;
    }
    break;

  case CODE_RIGHT:
    {
      pdmState = STATE_PATTERN_NEXT;
      interruptCondition = true;
    }
    break;
  }

  // User interaction detected so terminate pattern display
  if (interruptCondition) {
    return true;
  }

  // Now check timeout
  if (timeOutEnabled && (millis() > timeOut)) {
    // Timeout occurred
    timeOutEnabled = false;

    pdmState = STATE_PATTERN_SELECT;

    return true;
  }
  // Otherwise not time for pattern to exit
  return false;
}

// Pattern Display Mode function
void patternDisplayMode() {

  boolean maintainPattern = false;

  // Set initial state
  pdmState = STATE_PATTERN_SELECT;

  while (true) {

    switch (pdmState) {

    case STATE_PATTERN_SELECT:
      {
        if (! maintainPattern) {
          // Randomly select display pattern and speed
          selectorIndex = random(0, NUMBER_OF_PATTERNS);
          multiplierIndex = random(MIN_MULTIPLIER_INDEX, MAX_MULTIPLIER_INDEX + 1);
          calculateDelayMultiplier(multiplierIndex);
        }

        // Set next state
        pdmState = STATE_PATTERN_EXECUTE;
      }
      break;

    case STATE_PATTERN_EXECUTE:
      {
        // Turn off all LEDs
        clearLEDs();
        delay(300);

        // Calculate future time to switch pattern
        timeOut = millis() + (1000L * PATTERN_DURATION_SECONDS);

        // Enable time outs
        timeOutEnabled = true;

        // Start up the selected pattern
        (*patternFunctions[selectorIndex])();
      }
      break;

    case STATE_SPEED_UP:
      {
        multiplierIndex--;
        if (multiplierIndex < MIN_MULTIPLIER_INDEX) {
          multiplierIndex = MIN_MULTIPLIER_INDEX;
        }
        calculateDelayMultiplier(multiplierIndex);

        // Set next state
        pdmState = STATE_WAITING;
      }
      break;

    case STATE_SPEED_DOWN:
      {
        multiplierIndex++;
        if (multiplierIndex > MAX_MULTIPLIER_INDEX) {
          multiplierIndex = MAX_MULTIPLIER_INDEX;
        }
        calculateDelayMultiplier(multiplierIndex);

        // Set next state
        pdmState = STATE_WAITING;
      }
      break;

    case STATE_PATTERN_PREV:
      {
        selectorIndex--;
        if (selectorIndex < 0) {
          selectorIndex = NUMBER_OF_PATTERNS - 1;
        }

        // Set next state
        pdmState = STATE_WAITING;
      }
      break;

    case STATE_PATTERN_NEXT:
      {
        selectorIndex++;
        if (selectorIndex > NUMBER_OF_PATTERNS - 1) {
          selectorIndex = 0;
        }

        // Set next state
        pdmState = STATE_WAITING;
      }
      break;

    case STATE_WAITING:
      {
        irCode = readIRCode();

        // Process the code
        switch (irCode) {

        case CODE_ENTERSAVE:
          {
            // Pattern manually selected so maintain it
            maintainPattern = true;

            // Set next state
            pdmState = STATE_PATTERN_EXECUTE;
          }
          break;

        case CODE_UP:
          {
            // Set next state
            pdmState = STATE_SPEED_UP;
          }
          break;

        case CODE_DOWN:
          {
            // Set next state
            pdmState = STATE_SPEED_DOWN;
          }
          break;

        case CODE_LEFT:
          {
            // Set next state
            pdmState = STATE_PATTERN_PREV;
          }
          break;

        case CODE_RIGHT:
          {
            // Set next state
            pdmState = STATE_PATTERN_NEXT;
          }
          break;
        }
        break;
      }

    case STATE_EXITING:
      return;
    }
  }
}

/************************************************************************/
/************************** Clock Display Mode **************************/
/************************************************************************/

// This array maps which LEDs represent which hour. In the vertical clock
// the bottom LED (which is physical LED #2) represents 1:00 and the top
// LED (physical LED #14) represents 12:00. If you make a circular clock 
// you may want to change this mapping.

// Valid hours are 1 .. 12
uint8_t HOUR_LED_MAP [] = {
  2, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13
};


// Given the address in EEPROM of a HSV color entry, return the packed
// color it represents.
uint32_t colorFromEEPROMAddress(int address) {

  float hue, sat, val;
  struct RGB color;  

  hue = eepromReadFloat(address);
  sat = eepromReadFloat(address + 4);
  val = eepromReadFloat(address + 8);

  createColor(hue, sat, val, &color);
  return createPackedColor(color.red, color.green, color.blue);
}

// Flash the display to get user's attention
void attentionFlash() {

  clearLEDs();

  // Indicate we are in main state mode
  for (int i = 2; i < NUMBER_OF_LEDS; i++) {
    setLEDColor(i, 32, 32, 32);
  }
  show();

  delay(400);
  clearLEDs();
}

void clockDisplayMode() {

  // Clock operational states
  enum CLOCK_STATE {
    STATE_DISPATCH, STATE_RUNNING, STATE_EVENT10,
    STATE_EVENT15, STATE_EVENT30
  };

  // Retrieve colors from eeprom
  uint32_t backgroundColor = colorFromEEPROMAddress(CLK_BACKGROUND_HUE_ADDRESS);
  uint32_t minColor = colorFromEEPROMAddress(CLK_MIN_HUE_ADDRESS);
  uint32_t hourColor = colorFromEEPROMAddress(CLK_HOUR_HUE_ADDRESS);
  uint32_t dayColor = colorFromEEPROMAddress(CLK_DAY_HUE_ADDRESS);
  uint32_t monColor = colorFromEEPROMAddress(CLK_MON_HUE_ADDRESS);
  uint32_t yearColor = colorFromEEPROMAddress(CLK_YEAR_HUE_ADDRESS);
  uint32_t toggleColor = colorFromEEPROMAddress(CLK_TOGGLE_HUE_ADDRESS);

  uint32_t blackColor = createPackedColor(0, 0, 0);

  // Check for validity of the colors
  if (backgroundColor == 0) {
    // Set background color to blue
    backgroundColor = DEFAULT_BACKGROUND_COLOR;
  }

  if (minColor == 0) {
    // Set minute color to green
    minColor = DEFAULT_MIN_COLOR;
  }

  if (hourColor == 0) {
    // Set hour color to red
    hourColor = DEFAULT_HOUR_COLOR;
  }

  if (dayColor == 0) {
    // Set day color to magenta
    dayColor = DEFAULT_DAY_COLOR;
  }

  if (monColor == 0) {
    // Set month color to yellow
    monColor = DEFAULT_MON_COLOR;
  }

  if (yearColor == 0) {
    // Set year color to cyan
    yearColor = DEFAULT_YEAR_COLOR;
  }

  if (toggleColor == 0) {
    // Set toggle color to red
    toggleColor = DEFAULT_TOGGLE_COLOR;
  }

  // Set initial state of clock
  enum CLOCK_STATE clockState = STATE_DISPATCH;

  while (true) {

    // Get the current time and date from the RTC
    getTime();
    getDate();

    if (DEBUG) {
      printDateTime();
    }

    // Clock runs as state machine
    switch (clockState) {

    case STATE_DISPATCH:
      {
        // Read the remote
        irCode = readIRCode();

        // Has user changed mode ?
        if (irCode == CODE_STOPMODE) {
          return;
        }

        // Dispatch events in priority order
        // 30 minute event ? 
        if ((minutes != latchedMinute) && ((minutes % 30) == 0)) {
          clockState = STATE_EVENT30;
        }
        // 15 minute event ? 
        else if ((minutes != latchedMinute) && (minutes != 0) && ((minutes % 15) == 0)) {
          clockState = STATE_EVENT15;
        }
        // 10 minute event ? 
        else if ((minutes != latchedMinute) && (minutes != 0) && ((minutes % 10) == 0)) {
          clockState = STATE_EVENT10;
        }  
        else  {
          clockState = STATE_RUNNING;
        }
        break;
      }

    case STATE_EVENT10:
      {
        latchedMinute = minutes;

        // Every event do the rotating rainbow pattern
        struct RGB color;
        double hue;

        clearLEDs();

        // Hue increment
        double hueInc = 360.0 / 12.0;

        // Repeat the pattern specified times
        for (int i = 0; i < 8; i++) {
          // Pattern rotates 12 times so each color has been in each position
          for (int s = 0; s < 12; s++) {
            for (int led = 2; led < NUMBER_OF_LEDS; led++) {
              hue = ((int)((led + s) * hueInc)) % 360;
              createColor(hue, 1.0, 0.5, &color);
              setLEDColor(led, color);
            }
            show();    
            delay(250);
          } 
        }

        // Set next state
        clockState = STATE_DISPATCH;
        break;
      }  

    case STATE_EVENT15:
      {
        latchedMinute = minutes;

        // Every event display brightening blue display
        struct RGB color;

        clearLEDs();

        //Gradually increase blue value
        for (int s = 0; s < sizeof(GAMMA); s++) {
          float value = GAMMA[s] / 127.0;

          for (int led = 2; led < NUMBER_OF_LEDS; led++) {
            createColor(240, 1.0, value, &color);
            setLEDColor(led, color);
          }
          show();    
          delay(200);
        }

        clockState = STATE_DISPATCH;
        break;
      }

    case STATE_EVENT30:
      {
        latchedMinute = minutes;

        // Every event display date
        clearLEDs();

        // Display month 1..12
        setLEDColor(HOUR_LED_MAP[month], monColor);
        show();
        delay(3000);

        clearLEDs();

        // Display day 1..31
        int dayCount = date;
        int led = 2;

        while (--dayCount >= 0) {

          setLEDColor(led, dayColor);
          show();
          delay(400);
          setLEDColor(led, blackColor);
          show();

          led++;
          if (led >= NUMBER_OF_LEDS) {
            led = 2;
          }
        }
        delay(3000);

        // Display year 2010 - 2021
        clearLEDs();

        int yearOffset = year - 9;
        if (year < 0) {
          yearOffset = 4;  // 2013
        }

        setLEDColor(HOUR_LED_MAP[yearOffset], yearColor);
        show();
        delay(3000);

        // Set next state
        clockState = STATE_DISPATCH;
        break;
      }  

    case STATE_RUNNING:
      {
        // Display mode of operation

        // Toggle seconds indication
        if (secondsToggle) {
          setLEDColor(0, toggleColor);
          setLEDColor(1, blackColor);
          secondsToggle = false;
        }  
        else  {
          setLEDColor(0, blackColor);
          setLEDColor(1, toggleColor);
          secondsToggle = true;
        }
        show();

        // Set background color of clock
        for (int i = 2; i < NUMBER_OF_LEDS; i++) {
          setLEDColor(i, backgroundColor);
        }
        show();

        // Display the minute value as closest 5 minute LED
        uint8_t displayMinute = (uint8_t) minutes / 5;

        // Hour and Minute on same time ?
        // If so alternate hour and minute color on LED
        if (displayMinute == hours) {

          if (hourMinuteToggle) {
            hourMinuteToggle = false;
            setLEDColor(HOUR_LED_MAP[hours], hourColor);
          }  
          else  {
            hourMinuteToggle = true;
            setLEDColor(HOUR_LED_MAP[displayMinute], minColor);
          }
        }  
        else  {
          // Hour and minutes display different
          hourMinuteToggle = false;
          setLEDColor(HOUR_LED_MAP[hours], hourColor);
          setLEDColor(HOUR_LED_MAP[displayMinute], minColor);
        }
        show();
      }
      delay(1000);

      // Set next state
      clockState = STATE_DISPATCH;
      break; 
    } 
  }
}

/************************************************************************/
/****************************** Set Clock Time **************************/
/************************************************************************/

void setClockTime() {

  int setHour = 1;
  int set10Mins = 0;
  int setMin = 0;

  // Read colors for hours and mins
  uint32_t hourColor = colorFromEEPROMAddress(CLK_HOUR_HUE_ADDRESS);
  uint32_t minColor = colorFromEEPROMAddress(CLK_MIN_HUE_ADDRESS);

  if (hourColor == 0) {
    // Set hour color to default
    hourColor = DEFAULT_HOUR_COLOR;
  }

  if (minColor == 0) {
    // Set minute color to default
    minColor = DEFAULT_MIN_COLOR;
  }

  // Do attention flash
  attentionFlash();

  boolean done = false;

  // First we set the hour 1..12
  while (! done) {

    clearLEDs();

    // Set hour indicator
    setLEDColor(HOUR_LED_MAP[setHour], hourColor);
    show();

    // Read the remote
    irCode = readIRCode();

    while (irCode == 0) {
      delay(200);
      irCode = readIRCode();
    }

    switch (irCode) {
    case CODE_UP:
      {
        setHour++;
        if (setHour > 12) {
          setHour = 12;
        }
      }
      break;

    case CODE_DOWN:
      {
        setHour--;
        if (setHour < 1) {
          setHour = 1;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Set 10's of minutes

  attentionFlash();

  done = false;
  while (! done) {

    clearLEDs();

    // Conditionally set 10s of minutes indicator
    if (set10Mins > 0) {
      setLEDColor(HOUR_LED_MAP[set10Mins], minColor);
      show();
    }

    // Read the remote
    irCode = readIRCode();

    if (irCode == 0) {
      delay(100);
      continue;
    }

    switch (irCode) {
    case CODE_UP:
      {
        set10Mins++;
        if (set10Mins > 5) {
          set10Mins = 5;
        }
      }
      break;

    case CODE_DOWN:
      {
        set10Mins--;
        if (set10Mins < 0) {
          set10Mins = 0;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Set minutes

  attentionFlash();

  done = false;
  while (! done) {

    clearLEDs();

    // Conditionally set minutes indicator
    if (setMin > 0) {
      setLEDColor(HOUR_LED_MAP[setMin], minColor);
      show();
    }

    // Read the remote
    irCode = readIRCode();

    if (irCode == 0) {
      delay(100);
      continue;
    }

    switch (irCode) {
    case CODE_UP:
      {
        setMin++;
        if (setMin > 9) {
          setMin = 9;
        }
      }
      break;

    case CODE_DOWN:
      {
        setMin--;
        if (setMin < 0) {
          setMin = 0;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Now set the RTC
  hours = setHour;
  minutes = (set10Mins * 10) + setMin;
  seconds = 0;

  setTime();
}

/************************************************************************/
/****************************** Set Clock Date **************************/
/************************************************************************/

void setClockDate() {

  int setMon = 1;
  int set10Days = 0;
  int setDay = 1;
  int set10Years = 1;
  int setYear = 3;

  // Read colors for hours and mins
  uint32_t monColor = colorFromEEPROMAddress(CLK_MON_HUE_ADDRESS);
  uint32_t dayColor = colorFromEEPROMAddress(CLK_DAY_HUE_ADDRESS);
  uint32_t yearColor = colorFromEEPROMAddress(CLK_YEAR_HUE_ADDRESS);

  if (monColor == 0) {
    // Set month color to default
    monColor = DEFAULT_MON_COLOR;
  }

  if (dayColor == 0) {
    // Set day color to default
    dayColor = DEFAULT_DAY_COLOR;
  }

  if (yearColor == 0) {
    // Set year color to default
    yearColor = DEFAULT_YEAR_COLOR;
  }

  // Do attention flash
  attentionFlash();

  boolean done = false;

  // First we set the month 1..12
  while (! done) {

    clearLEDs();

    // Set month indicator
    setLEDColor(HOUR_LED_MAP[setMon], monColor);
    show();

    // Read the remote
    irCode = readIRCode();

    while (irCode == 0) {
      delay(200);
      irCode = readIRCode();
    }

    switch (irCode) {
    case CODE_UP:
      {
        setMon++;
        if (setMon > 12) {
          setMon = 12;
        }
      }
      break;

    case CODE_DOWN:
      {
        setMon--;
        if (setMon < 1) {
          setMon = 1;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Set 10's of days

  attentionFlash();

  done = false;
  while (! done) {

    clearLEDs();

    // Conditionally set 10s of days indicator
    if (set10Days > 0) {
      setLEDColor(HOUR_LED_MAP[set10Days], dayColor);
      show();
    }

    // Read the remote
    irCode = readIRCode();

    if (irCode == 0) {
      delay(100);
      continue;
    }

    switch (irCode) {
    case CODE_UP:
      {
        set10Days++;
        if (set10Days > 3) {
          set10Days = 3;
        }
      }
      break;

    case CODE_DOWN:
      {
        set10Days--;
        if (set10Days < 0) {
          set10Days = 0;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Set days

  attentionFlash();

  done = false;
  while (! done) {

    clearLEDs();

    // Conditionally set minutes indicator
    if (setDay > 0) {
      setLEDColor(HOUR_LED_MAP[setDay], dayColor);
      show();
    }

    // Read the remote
    irCode = readIRCode();

    if (irCode == 0) {
      delay(100);
      continue;
    }

    switch (irCode) {
    case CODE_UP:
      {
        setDay++;
        if (setDay > 9) {
          setDay = 9;
        }
      }
      break;

    case CODE_DOWN:
      {
        setDay--;
        if (setDay < 0) {
          setDay = 0;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Set 10's of years

  attentionFlash();

  done = false;
  while (! done) {

    clearLEDs();

    // Conditionally set 10s of years indicator
    setLEDColor(HOUR_LED_MAP[set10Years], yearColor);
    show();

    // Read the remote
    irCode = readIRCode();

    if (irCode == 0) {
      delay(100);
      continue;
    }

    switch (irCode) {
    case CODE_UP:
      {
        set10Years++;
        if (set10Years > 9) {
          set10Years = 9;
        }
      }
      break;

    case CODE_DOWN:
      {
        set10Years--;
        if (set10Years < 1) {
          set10Years = 1;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Set year

  attentionFlash();

  done = false;
  while (! done) {

    clearLEDs();

    // Conditionally set year indicator
    setLEDColor(HOUR_LED_MAP[setYear], yearColor);
    show();

    // Read the remote
    irCode = readIRCode();

    if (irCode == 0) {
      delay(100);
      continue;
    }

    switch (irCode) {
    case CODE_UP:
      {
        setYear++;
        if (setYear > 9) {
          setYear = 9;
        }
      }
      break;

    case CODE_DOWN:
      {
        setYear--;
        if (setYear < 0) {
          setYear = 0;
        }
      }
      break;

    case CODE_ENTERSAVE:
      {
        done = true;
      }
      break;
    }
  }

  // Now set the RTC
  month = setMon;
  date = (set10Days * 10) + setDay;
  year = (set10Years * 10) + setYear;

  setDate();
}

/************************************************************************/
/************************* Set Background Color *************************/
/************************************************************************/

// Set background color callback
void bgcCallback(struct HSV *hsv) {

  struct RGB rgbColor;

  // Create color from hsv data
  createColor(hsv->hue, hsv->sat, hsv->val, &rgbColor);

  // Set all non toggle LEDs
  for (int i = 2; i < NUMBER_OF_LEDS; i++) {
    setLEDColor(i, rgbColor);
  }
  show();
}

// Allow manual adjustment of background color
void setBackgroundColor() {

  struct HSV hsvColor;

  float hue, sat, val;

  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_BACKGROUND_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_BACKGROUND_SAT_ADDRESS);
  val = eepromReadFloat(CLK_BACKGROUND_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to blue
    hue = 240.0;
    sat = 1.0;
    val = 0.5;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &bgcCallback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_BACKGROUND_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_BACKGROUND_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_BACKGROUND_VAL_ADDRESS, hsvColor.val);
  }
}

/************************************************************************/
/**************************** Set Time Colors ***************************/
/************************************************************************/

// Set time color callback
void tc2Callback(struct HSV *hsv) {

  struct RGB rgbColor;

  createColor(hsv->hue, hsv->sat, hsv->val, &rgbColor);
  setLEDColor(NUMBER_OF_LEDS - 1, rgbColor);
  show();
}

// Allow manual adjustment of time colors
void setTimeColors() {

  struct HSV hsvColor;

  float hue, sat, val;

  // Set the hour color
  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_HOUR_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_HOUR_SAT_ADDRESS);
  val = eepromReadFloat(CLK_HOUR_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to red
    hue = 0.0;
    sat = 1.0;
    val = 0.5;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &tc2Callback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_HOUR_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_HOUR_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_HOUR_VAL_ADDRESS, hsvColor.val);
  }

  // Set the minute color
  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_MIN_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_MIN_SAT_ADDRESS);
  val = eepromReadFloat(CLK_MIN_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to green
    hue = 120.0;
    sat = 1.0;
    val = 0.5;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &tc2Callback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_MIN_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_MIN_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_MIN_VAL_ADDRESS, hsvColor.val);
  }
}

/************************************************************************/
/**************************** Set Date Colors ***************************/
/************************************************************************/

// Set date color callback
void dcCallback(struct HSV *hsv) {

  struct RGB rgbColor;

  createColor(hsv->hue, hsv->sat, hsv->val, &rgbColor);
  setLEDColor(NUMBER_OF_LEDS - 1, rgbColor);
  show();
}

// Allow manual adjustment of date colors
void setDateColors() {

  struct HSV hsvColor;

  float hue, sat, val;

  // Set the month color
  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_MON_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_MON_SAT_ADDRESS);
  val = eepromReadFloat(CLK_MON_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to yellow
    hue = 50.0;
    sat = 1.0;
    val = 0.5;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &dcCallback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_MON_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_MON_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_MON_VAL_ADDRESS, hsvColor.val);
  }

  // Set the day color
  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_DAY_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_DAY_SAT_ADDRESS);
  val = eepromReadFloat(CLK_DAY_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to magenta
    hue = 329.0;
    sat =  0.66;
    val =  0.97;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &dcCallback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_DAY_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_DAY_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_DAY_VAL_ADDRESS, hsvColor.val);
  }

  // Set the year color
  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_YEAR_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_YEAR_SAT_ADDRESS);
  val = eepromReadFloat(CLK_YEAR_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to cyan
    hue = 180.0;
    sat =   1.0;
    val =   0.97;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &dcCallback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_YEAR_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_YEAR_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_YEAR_VAL_ADDRESS, hsvColor.val);
  }
}

/************************************************************************/
/*************************** Set Toggle Color ***************************/
/************************************************************************/

// Set toggle color callback
void tcCallback(struct HSV *hsv) {

  struct RGB rgbColor;

  createColor(hsv->hue, hsv->sat, hsv->val, &rgbColor);
  for (int i = 0; i < 2; i++) {
    setLEDColor(i, rgbColor);
  }
  show();
}

// Allow manual adjustment of toggle color
void setToggleColor() {

  struct HSV hsvColor;

  float hue, sat, val;

  // Read the values from the eeprom
  hue = eepromReadFloat(CLK_TOGGLE_HUE_ADDRESS);
  sat = eepromReadFloat(CLK_TOGGLE_SAT_ADDRESS);
  val = eepromReadFloat(CLK_TOGGLE_VAL_ADDRESS);

  // Has a previous value been set ?
  if ((hue == 0.0) && (sat == 0.0) && (val == 0.0)) {
    // No previous value so set default to red
    hue = 0.0;
    sat = 1.0;
    val = 0.5;
  }

  // Load up structure
  hsvColor.hue = hue;
  hsvColor.sat = sat;
  hsvColor.val = val;

  // Call remote control code for setting hsv values
  hsvColorSetter(&hsvColor, &tcCallback);

  // Save color if changed by user
  if ((hsvColor.hue != hue) || (hsvColor.sat != sat) || (hsvColor.val != val)) {
    eepromWriteFloat(CLK_TOGGLE_HUE_ADDRESS, hsvColor.hue);
    eepromWriteFloat(CLK_TOGGLE_SAT_ADDRESS, hsvColor.sat);
    eepromWriteFloat(CLK_TOGGLE_VAL_ADDRESS, hsvColor.val);
  }
}

/************************************************************************/
/********** Setup all subsystems before clock starts operation **********/
/************************************************************************/

void setup() {

  if (DEBUG) {
    // Setup serial link for debugging
    Serial.begin(115200);
  }

  // As analog input pin 0 is unconnected, random analog
  // noise will cause the call to randomSeed() to generate
  // different seed numbers each time the sketch runs.
  // randomSeed() will then shuffle the random function.
  randomSeed(analogRead(0));

  // Start the IR receiver
  irReceiver.enableIRIn();

  // Turn off onboard yellow LED
  pinMode(LED_PIN, INPUT);
  digitalWrite(LED_PIN, HIGH);

  // Setup LPD8806 clock and data pins as output
  pinMode(LED_STRIP_CLOCK_PIN, OUTPUT);
  digitalWrite(LED_STRIP_CLOCK_PIN, LOW);
  pinMode(LED_STRIP_DATA_PIN, OUTPUT);
  digitalWrite(LED_STRIP_DATA_PIN, LOW);

  // Establish masks for quick access to LED strip
  clockPort = portOutputRegister(digitalPinToPort(LED_STRIP_CLOCK_PIN));
  clockPinMask = digitalPinToBitMask(LED_STRIP_CLOCK_PIN);
  dataPort = portOutputRegister(digitalPinToPort(LED_STRIP_DATA_PIN));
  dataPinMask = digitalPinToBitMask(LED_STRIP_DATA_PIN);

  // Set all LEDs in strip to black color
  clearLEDs();

  secondsToggle = false;
  hourMinuteToggle = false;

  // Startup Wire library
  Wire.begin();

  // Check for RTC operation
  if (! rtcRunning()) {
    if (DEBUG) {
      Serial.println("RTC is NOT running!");
    }
  }

  // Put RTC into 12 hour format
  set12HourFormat();

  // Determine if eeprom has been initialized by looking for signature
  uint8_t sig0 = eepromReadUint8_t(0);
  uint8_t sig1 = eepromReadUint8_t(1);
  uint8_t sig2 = eepromReadUint8_t(2);
  uint8_t sig3 = eepromReadUint8_t(3);

  if ((sig0 != 0xAA) | (sig1 != 0x00) | (sig2 != 0xFF) | (sig3 != 0x55)) {

    // EEPROM is uninitialized. Initialize signature and all hsv color values
    resetEEPROMData();
  }
}

// Program entry point
void loop() {    

  enum STATES {
    STATE_WAIT, STATE_CLOCK_MODE, STATE_MOODLIGHT_MODE, STATE_PATTERN_MODE,
    STATE_SET_TIME, STATE_SET_DATE, STATE_SET_BACKGROUND_COLOR,
    STATE_SET_TIME_COLORS, STATE_SET_DATE_COLORS, STATE_SET_TOGGLE_COLOR,
    STATE_EEPROM_RESET
  };

  boolean flashed = false;

  enum STATES mainState = STATE_CLOCK_MODE;

  while (true) {

    switch (mainState) {

    case STATE_WAIT:
      {
        if (! flashed) {
          flashed = true;

          attentionFlash();
        }

        // Read the remote
        irCode = readIRCode();

        switch (irCode) {

        case CODE_1:
          // Set next state
          mainState = STATE_CLOCK_MODE;
          break;

        case CODE_2:
          // Set next state
          mainState = STATE_MOODLIGHT_MODE;
          break;

        case CODE_3:
          // Set next state
          mainState = STATE_PATTERN_MODE;
          break;

        case CODE_4:
          // Set next state
          mainState = STATE_SET_TIME;
          break;

        case CODE_5:
          // Set next state
          mainState = STATE_SET_DATE;
          break;

        case CODE_6:
          // Set next state
          mainState = STATE_SET_BACKGROUND_COLOR;
          break;

        case CODE_7:
          // Set next state
          mainState = STATE_SET_TIME_COLORS;
          break;

        case CODE_8:
          // Set next state
          mainState = STATE_SET_DATE_COLORS;
          break;

        case CODE_9:
          // Set next state
          mainState = STATE_SET_TOGGLE_COLOR;
          break;

        case CODE_SETUP:
          // Set next state
          mainState = STATE_EEPROM_RESET;
          break;
        }            
      }
      break;

    case STATE_CLOCK_MODE:
      {
        clockDisplayMode();

        // Set next state
        mainState = STATE_WAIT;
        flashed = false;
      }
      break;

    case STATE_MOODLIGHT_MODE:
      {
        moodLightMode();

        // Set next state
        mainState = STATE_WAIT;
        flashed = false;
      }
      break;

    case STATE_PATTERN_MODE:
      {
        patternDisplayMode();

        // Set next state
        mainState = STATE_WAIT;
        flashed = false;
      }
      break;

    case STATE_SET_TIME:
      {
        setClockTime();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;

    case STATE_SET_DATE:
      {
        setClockDate();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;

    case STATE_SET_BACKGROUND_COLOR:
      {
        setBackgroundColor();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;

    case STATE_SET_TIME_COLORS:
      {
        setTimeColors();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;

    case STATE_SET_DATE_COLORS:
      {
        setDateColors();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;

    case STATE_SET_TOGGLE_COLOR:
      {
        setToggleColor();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;

    case STATE_EEPROM_RESET:
      {
        resetEEPROMData();

        // Set next state
        mainState = STATE_CLOCK_MODE;
        flashed = false;
      }
      break;
    }
  }
}















































